# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cisco_telemetry.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='cisco_telemetry.proto',
  package='telemetry',
  syntax='proto3',
  serialized_pb=_b('\n\x15\x63isco_telemetry.proto\x12\ttelemetry\"\xbc\x02\n\tTelemetry\x12\x15\n\x0bnode_id_str\x18\x01 \x01(\tH\x00\x12\x1d\n\x13subscription_id_str\x18\x03 \x01(\tH\x01\x12\x15\n\rencoding_path\x18\x06 \x01(\t\x12\x15\n\rcollection_id\x18\x08 \x01(\x04\x12\x1d\n\x15\x63ollection_start_time\x18\t \x01(\x04\x12\x15\n\rmsg_timestamp\x18\n \x01(\x04\x12-\n\ndata_gpbkv\x18\x0b \x03(\x0b\x32\x19.telemetry.TelemetryField\x12.\n\x08\x64\x61ta_gpb\x18\x0c \x01(\x0b\x32\x1c.telemetry.TelemetryGPBTable\x12\x1b\n\x13\x63ollection_end_time\x18\r \x01(\x04\x42\t\n\x07node_idB\x0e\n\x0csubscription\"\xc1\x02\n\x0eTelemetryField\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x15\n\x0b\x62ytes_value\x18\x04 \x01(\x0cH\x00\x12\x16\n\x0cstring_value\x18\x05 \x01(\tH\x00\x12\x14\n\nbool_value\x18\x06 \x01(\x08H\x00\x12\x16\n\x0cuint32_value\x18\x07 \x01(\rH\x00\x12\x16\n\x0cuint64_value\x18\x08 \x01(\x04H\x00\x12\x16\n\x0csint32_value\x18\t \x01(\x11H\x00\x12\x16\n\x0csint64_value\x18\n \x01(\x12H\x00\x12\x16\n\x0c\x64ouble_value\x18\x0b \x01(\x01H\x00\x12\x15\n\x0b\x66loat_value\x18\x0c \x01(\x02H\x00\x12)\n\x06\x66ields\x18\x0f \x03(\x0b\x32\x19.telemetry.TelemetryFieldB\x0f\n\rvalue_by_type\"<\n\x11TelemetryGPBTable\x12\'\n\x03row\x18\x01 \x03(\x0b\x32\x1a.telemetry.TelemetryRowGPB\"C\n\x0fTelemetryRowGPB\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\x12\x0c\n\x04keys\x18\n \x01(\x0c\x12\x0f\n\x07\x63ontent\x18\x0b \x01(\x0c\x62\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_TELEMETRY = _descriptor.Descriptor(
  name='Telemetry',
  full_name='telemetry.Telemetry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='node_id_str', full_name='telemetry.Telemetry.node_id_str', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='subscription_id_str', full_name='telemetry.Telemetry.subscription_id_str', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='encoding_path', full_name='telemetry.Telemetry.encoding_path', index=2,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='collection_id', full_name='telemetry.Telemetry.collection_id', index=3,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='collection_start_time', full_name='telemetry.Telemetry.collection_start_time', index=4,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='msg_timestamp', full_name='telemetry.Telemetry.msg_timestamp', index=5,
      number=10, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='data_gpbkv', full_name='telemetry.Telemetry.data_gpbkv', index=6,
      number=11, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='data_gpb', full_name='telemetry.Telemetry.data_gpb', index=7,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='collection_end_time', full_name='telemetry.Telemetry.collection_end_time', index=8,
      number=13, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='node_id', full_name='telemetry.Telemetry.node_id',
      index=0, containing_type=None, fields=[]),
    _descriptor.OneofDescriptor(
      name='subscription', full_name='telemetry.Telemetry.subscription',
      index=1, containing_type=None, fields=[]),
  ],
  serialized_start=37,
  serialized_end=353,
)


_TELEMETRYFIELD = _descriptor.Descriptor(
  name='TelemetryField',
  full_name='telemetry.TelemetryField',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='telemetry.TelemetryField.timestamp', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='telemetry.TelemetryField.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='bytes_value', full_name='telemetry.TelemetryField.bytes_value', index=2,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='string_value', full_name='telemetry.TelemetryField.string_value', index=3,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='bool_value', full_name='telemetry.TelemetryField.bool_value', index=4,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='uint32_value', full_name='telemetry.TelemetryField.uint32_value', index=5,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='uint64_value', full_name='telemetry.TelemetryField.uint64_value', index=6,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sint32_value', full_name='telemetry.TelemetryField.sint32_value', index=7,
      number=9, type=17, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sint64_value', full_name='telemetry.TelemetryField.sint64_value', index=8,
      number=10, type=18, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='double_value', full_name='telemetry.TelemetryField.double_value', index=9,
      number=11, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='float_value', full_name='telemetry.TelemetryField.float_value', index=10,
      number=12, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fields', full_name='telemetry.TelemetryField.fields', index=11,
      number=15, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value_by_type', full_name='telemetry.TelemetryField.value_by_type',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=356,
  serialized_end=677,
)


_TELEMETRYGPBTABLE = _descriptor.Descriptor(
  name='TelemetryGPBTable',
  full_name='telemetry.TelemetryGPBTable',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='row', full_name='telemetry.TelemetryGPBTable.row', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=679,
  serialized_end=739,
)


_TELEMETRYROWGPB = _descriptor.Descriptor(
  name='TelemetryRowGPB',
  full_name='telemetry.TelemetryRowGPB',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='telemetry.TelemetryRowGPB.timestamp', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='keys', full_name='telemetry.TelemetryRowGPB.keys', index=1,
      number=10, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='content', full_name='telemetry.TelemetryRowGPB.content', index=2,
      number=11, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=741,
  serialized_end=808,
)

_TELEMETRY.fields_by_name['data_gpbkv'].message_type = _TELEMETRYFIELD
_TELEMETRY.fields_by_name['data_gpb'].message_type = _TELEMETRYGPBTABLE
_TELEMETRY.oneofs_by_name['node_id'].fields.append(
  _TELEMETRY.fields_by_name['node_id_str'])
_TELEMETRY.fields_by_name['node_id_str'].containing_oneof = _TELEMETRY.oneofs_by_name['node_id']
_TELEMETRY.oneofs_by_name['subscription'].fields.append(
  _TELEMETRY.fields_by_name['subscription_id_str'])
_TELEMETRY.fields_by_name['subscription_id_str'].containing_oneof = _TELEMETRY.oneofs_by_name['subscription']
_TELEMETRYFIELD.fields_by_name['fields'].message_type = _TELEMETRYFIELD
_TELEMETRYFIELD.oneofs_by_name['value_by_type'].fields.append(
  _TELEMETRYFIELD.fields_by_name['bytes_value'])
_TELEMETRYFIELD.fields_by_name['bytes_value'].containing_oneof = _TELEMETRYFIELD.oneofs_by_name['value_by_type']
_TELEMETRYFIELD.oneofs_by_name['value_by_type'].fields.append(
  _TELEMETRYFIELD.fields_by_name['string_value'])
_TELEMETRYFIELD.fields_by_name['string_value'].containing_oneof = _TELEMETRYFIELD.oneofs_by_name['value_by_type']
_TELEMETRYFIELD.oneofs_by_name['value_by_type'].fields.append(
  _TELEMETRYFIELD.fields_by_name['bool_value'])
_TELEMETRYFIELD.fields_by_name['bool_value'].containing_oneof = _TELEMETRYFIELD.oneofs_by_name['value_by_type']
_TELEMETRYFIELD.oneofs_by_name['value_by_type'].fields.append(
  _TELEMETRYFIELD.fields_by_name['uint32_value'])
_TELEMETRYFIELD.fields_by_name['uint32_value'].containing_oneof = _TELEMETRYFIELD.oneofs_by_name['value_by_type']
_TELEMETRYFIELD.oneofs_by_name['value_by_type'].fields.append(
  _TELEMETRYFIELD.fields_by_name['uint64_value'])
_TELEMETRYFIELD.fields_by_name['uint64_value'].containing_oneof = _TELEMETRYFIELD.oneofs_by_name['value_by_type']
_TELEMETRYFIELD.oneofs_by_name['value_by_type'].fields.append(
  _TELEMETRYFIELD.fields_by_name['sint32_value'])
_TELEMETRYFIELD.fields_by_name['sint32_value'].containing_oneof = _TELEMETRYFIELD.oneofs_by_name['value_by_type']
_TELEMETRYFIELD.oneofs_by_name['value_by_type'].fields.append(
  _TELEMETRYFIELD.fields_by_name['sint64_value'])
_TELEMETRYFIELD.fields_by_name['sint64_value'].containing_oneof = _TELEMETRYFIELD.oneofs_by_name['value_by_type']
_TELEMETRYFIELD.oneofs_by_name['value_by_type'].fields.append(
  _TELEMETRYFIELD.fields_by_name['double_value'])
_TELEMETRYFIELD.fields_by_name['double_value'].containing_oneof = _TELEMETRYFIELD.oneofs_by_name['value_by_type']
_TELEMETRYFIELD.oneofs_by_name['value_by_type'].fields.append(
  _TELEMETRYFIELD.fields_by_name['float_value'])
_TELEMETRYFIELD.fields_by_name['float_value'].containing_oneof = _TELEMETRYFIELD.oneofs_by_name['value_by_type']
_TELEMETRYGPBTABLE.fields_by_name['row'].message_type = _TELEMETRYROWGPB
DESCRIPTOR.message_types_by_name['Telemetry'] = _TELEMETRY
DESCRIPTOR.message_types_by_name['TelemetryField'] = _TELEMETRYFIELD
DESCRIPTOR.message_types_by_name['TelemetryGPBTable'] = _TELEMETRYGPBTABLE
DESCRIPTOR.message_types_by_name['TelemetryRowGPB'] = _TELEMETRYROWGPB

Telemetry = _reflection.GeneratedProtocolMessageType('Telemetry', (_message.Message,), dict(
  DESCRIPTOR = _TELEMETRY,
  __module__ = 'cisco_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:telemetry.Telemetry)
  ))
_sym_db.RegisterMessage(Telemetry)

TelemetryField = _reflection.GeneratedProtocolMessageType('TelemetryField', (_message.Message,), dict(
  DESCRIPTOR = _TELEMETRYFIELD,
  __module__ = 'cisco_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:telemetry.TelemetryField)
  ))
_sym_db.RegisterMessage(TelemetryField)

TelemetryGPBTable = _reflection.GeneratedProtocolMessageType('TelemetryGPBTable', (_message.Message,), dict(
  DESCRIPTOR = _TELEMETRYGPBTABLE,
  __module__ = 'cisco_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:telemetry.TelemetryGPBTable)
  ))
_sym_db.RegisterMessage(TelemetryGPBTable)

TelemetryRowGPB = _reflection.GeneratedProtocolMessageType('TelemetryRowGPB', (_message.Message,), dict(
  DESCRIPTOR = _TELEMETRYROWGPB,
  __module__ = 'cisco_telemetry_pb2'
  # @@protoc_insertion_point(class_scope:telemetry.TelemetryRowGPB)
  ))
_sym_db.RegisterMessage(TelemetryRowGPB)


# @@protoc_insertion_point(module_scope)
