# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: huawei-ifm.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='huawei-ifm.proto',
  package='huawei_ifm',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x10huawei-ifm.proto\x12\nhuawei_ifm\"\xc1\x08\n\x03Ifm\x12.\n\ninterfaces\x18\x05 \x01(\x0b\x32\x1a.huawei_ifm.Ifm.Interfaces\x1a\x89\x08\n\nInterfaces\x12\x37\n\tinterface\x18\x01 \x03(\x0b\x32$.huawei_ifm.Ifm.Interfaces.Interface\x1a\xc1\x07\n\tInterface\x12I\n\rifAdminStatus\x18\x05 \x01(\x0e\x32\x32.huawei_ifm.Ifm.Interfaces.Interface.IfAdminStatus\x12I\n\rifClearedStat\x18\t \x01(\x0b\x32\x32.huawei_ifm.Ifm.Interfaces.Interface.IfClearedStat\x12I\n\rifDynamicInfo\x18\r \x01(\x0b\x32\x32.huawei_ifm.Ifm.Interfaces.Interface.IfDynamicInfo\x12\x0f\n\x07ifIndex\x18\x10 \x01(\r\x12\x0e\n\x06ifName\x18\x19 \x01(\t\x12G\n\x0cifStatistics\x18# \x01(\x0b\x32\x31.huawei_ifm.Ifm.Interfaces.Interface.IfStatistics\x1a\x35\n\rIfClearedStat\x12\x10\n\x08sendByte\x18\x1d \x01(\x04\x12\x12\n\nsendPacket\x18$ \x01(\x04\x1a\xa2\x01\n\rIfDynamicInfo\x12U\n\x0cifOperStatus\x18\x05 \x01(\x0e\x32?.huawei_ifm.Ifm.Interfaces.Interface.IfDynamicInfo.IfOperStatus\":\n\x0cIfOperStatus\x12\x15\n\x11IfOperStatus_DOWN\x10\x00\x12\x13\n\x0fIfOperStatus_UP\x10\x01\x1a\xcd\x02\n\x0cIfStatistics\x12\x16\n\x0ercvBroadPacket\x18\x02 \x01(\x04\x12\x15\n\rrcvDropPacket\x18\x03 \x01(\x04\x12\x16\n\x0ercvErrorPacket\x18\x04 \x01(\x04\x12\x15\n\rrcvMutiPacket\x18\x05 \x01(\x04\x12\x14\n\x0crcvUniPacket\x18\x06 \x01(\x04\x12\x13\n\x0breceiveByte\x18\x07 \x01(\x04\x12\x15\n\rreceivePacket\x18\x08 \x01(\x04\x12\x17\n\x0fsendBroadPacket\x18\t \x01(\x04\x12\x10\n\x08sendByte\x18\n \x01(\x04\x12\x16\n\x0esendDropPacket\x18\x0b \x01(\x04\x12\x17\n\x0fsendErrorPacket\x18\x0c \x01(\x04\x12\x16\n\x0esendMutiPacket\x18\r \x01(\x04\x12\x12\n\nsendPacket\x18\x0e \x01(\x04\x12\x15\n\rsendUniPacket\x18\x0f \x01(\x04\"=\n\rIfAdminStatus\x12\x16\n\x12IfAdminStatus_DOWN\x10\x00\x12\x14\n\x10IfAdminStatus_UP\x10\x01\x62\x06proto3')
)



_IFM_INTERFACES_INTERFACE_IFDYNAMICINFO_IFOPERSTATUS = _descriptor.EnumDescriptor(
  name='IfOperStatus',
  full_name='huawei_ifm.Ifm.Interfaces.Interface.IfDynamicInfo.IfOperStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='IfOperStatus_DOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IfOperStatus_UP', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=665,
  serialized_end=723,
)
_sym_db.RegisterEnumDescriptor(_IFM_INTERFACES_INTERFACE_IFDYNAMICINFO_IFOPERSTATUS)

_IFM_INTERFACES_INTERFACE_IFADMINSTATUS = _descriptor.EnumDescriptor(
  name='IfAdminStatus',
  full_name='huawei_ifm.Ifm.Interfaces.Interface.IfAdminStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='IfAdminStatus_DOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IfAdminStatus_UP', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1061,
  serialized_end=1122,
)
_sym_db.RegisterEnumDescriptor(_IFM_INTERFACES_INTERFACE_IFADMINSTATUS)


_IFM_INTERFACES_INTERFACE_IFCLEAREDSTAT = _descriptor.Descriptor(
  name='IfClearedStat',
  full_name='huawei_ifm.Ifm.Interfaces.Interface.IfClearedStat',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sendByte', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfClearedStat.sendByte', index=0,
      number=29, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sendPacket', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfClearedStat.sendPacket', index=1,
      number=36, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=505,
  serialized_end=558,
)

_IFM_INTERFACES_INTERFACE_IFDYNAMICINFO = _descriptor.Descriptor(
  name='IfDynamicInfo',
  full_name='huawei_ifm.Ifm.Interfaces.Interface.IfDynamicInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ifOperStatus', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfDynamicInfo.ifOperStatus', index=0,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _IFM_INTERFACES_INTERFACE_IFDYNAMICINFO_IFOPERSTATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=561,
  serialized_end=723,
)

_IFM_INTERFACES_INTERFACE_IFSTATISTICS = _descriptor.Descriptor(
  name='IfStatistics',
  full_name='huawei_ifm.Ifm.Interfaces.Interface.IfStatistics',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rcvBroadPacket', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfStatistics.rcvBroadPacket', index=0,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rcvDropPacket', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfStatistics.rcvDropPacket', index=1,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rcvErrorPacket', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfStatistics.rcvErrorPacket', index=2,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rcvMutiPacket', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfStatistics.rcvMutiPacket', index=3,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rcvUniPacket', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfStatistics.rcvUniPacket', index=4,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='receiveByte', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfStatistics.receiveByte', index=5,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='receivePacket', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfStatistics.receivePacket', index=6,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sendBroadPacket', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfStatistics.sendBroadPacket', index=7,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sendByte', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfStatistics.sendByte', index=8,
      number=10, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sendDropPacket', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfStatistics.sendDropPacket', index=9,
      number=11, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sendErrorPacket', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfStatistics.sendErrorPacket', index=10,
      number=12, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sendMutiPacket', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfStatistics.sendMutiPacket', index=11,
      number=13, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sendPacket', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfStatistics.sendPacket', index=12,
      number=14, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sendUniPacket', full_name='huawei_ifm.Ifm.Interfaces.Interface.IfStatistics.sendUniPacket', index=13,
      number=15, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=726,
  serialized_end=1059,
)

_IFM_INTERFACES_INTERFACE = _descriptor.Descriptor(
  name='Interface',
  full_name='huawei_ifm.Ifm.Interfaces.Interface',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ifAdminStatus', full_name='huawei_ifm.Ifm.Interfaces.Interface.ifAdminStatus', index=0,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ifClearedStat', full_name='huawei_ifm.Ifm.Interfaces.Interface.ifClearedStat', index=1,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ifDynamicInfo', full_name='huawei_ifm.Ifm.Interfaces.Interface.ifDynamicInfo', index=2,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ifIndex', full_name='huawei_ifm.Ifm.Interfaces.Interface.ifIndex', index=3,
      number=16, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ifName', full_name='huawei_ifm.Ifm.Interfaces.Interface.ifName', index=4,
      number=25, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ifStatistics', full_name='huawei_ifm.Ifm.Interfaces.Interface.ifStatistics', index=5,
      number=35, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_IFM_INTERFACES_INTERFACE_IFCLEAREDSTAT, _IFM_INTERFACES_INTERFACE_IFDYNAMICINFO, _IFM_INTERFACES_INTERFACE_IFSTATISTICS, ],
  enum_types=[
    _IFM_INTERFACES_INTERFACE_IFADMINSTATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=161,
  serialized_end=1122,
)

_IFM_INTERFACES = _descriptor.Descriptor(
  name='Interfaces',
  full_name='huawei_ifm.Ifm.Interfaces',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='interface', full_name='huawei_ifm.Ifm.Interfaces.interface', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_IFM_INTERFACES_INTERFACE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=89,
  serialized_end=1122,
)

_IFM = _descriptor.Descriptor(
  name='Ifm',
  full_name='huawei_ifm.Ifm',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='interfaces', full_name='huawei_ifm.Ifm.interfaces', index=0,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_IFM_INTERFACES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=33,
  serialized_end=1122,
)

_IFM_INTERFACES_INTERFACE_IFCLEAREDSTAT.containing_type = _IFM_INTERFACES_INTERFACE
_IFM_INTERFACES_INTERFACE_IFDYNAMICINFO.fields_by_name['ifOperStatus'].enum_type = _IFM_INTERFACES_INTERFACE_IFDYNAMICINFO_IFOPERSTATUS
_IFM_INTERFACES_INTERFACE_IFDYNAMICINFO.containing_type = _IFM_INTERFACES_INTERFACE
_IFM_INTERFACES_INTERFACE_IFDYNAMICINFO_IFOPERSTATUS.containing_type = _IFM_INTERFACES_INTERFACE_IFDYNAMICINFO
_IFM_INTERFACES_INTERFACE_IFSTATISTICS.containing_type = _IFM_INTERFACES_INTERFACE
_IFM_INTERFACES_INTERFACE.fields_by_name['ifAdminStatus'].enum_type = _IFM_INTERFACES_INTERFACE_IFADMINSTATUS
_IFM_INTERFACES_INTERFACE.fields_by_name['ifClearedStat'].message_type = _IFM_INTERFACES_INTERFACE_IFCLEAREDSTAT
_IFM_INTERFACES_INTERFACE.fields_by_name['ifDynamicInfo'].message_type = _IFM_INTERFACES_INTERFACE_IFDYNAMICINFO
_IFM_INTERFACES_INTERFACE.fields_by_name['ifStatistics'].message_type = _IFM_INTERFACES_INTERFACE_IFSTATISTICS
_IFM_INTERFACES_INTERFACE.containing_type = _IFM_INTERFACES
_IFM_INTERFACES_INTERFACE_IFADMINSTATUS.containing_type = _IFM_INTERFACES_INTERFACE
_IFM_INTERFACES.fields_by_name['interface'].message_type = _IFM_INTERFACES_INTERFACE
_IFM_INTERFACES.containing_type = _IFM
_IFM.fields_by_name['interfaces'].message_type = _IFM_INTERFACES
DESCRIPTOR.message_types_by_name['Ifm'] = _IFM
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Ifm = _reflection.GeneratedProtocolMessageType('Ifm', (_message.Message,), {

  'Interfaces' : _reflection.GeneratedProtocolMessageType('Interfaces', (_message.Message,), {

    'Interface' : _reflection.GeneratedProtocolMessageType('Interface', (_message.Message,), {

      'IfClearedStat' : _reflection.GeneratedProtocolMessageType('IfClearedStat', (_message.Message,), {
        'DESCRIPTOR' : _IFM_INTERFACES_INTERFACE_IFCLEAREDSTAT,
        '__module__' : 'huawei_ifm_pb2'
        # @@protoc_insertion_point(class_scope:huawei_ifm.Ifm.Interfaces.Interface.IfClearedStat)
        })
      ,

      'IfDynamicInfo' : _reflection.GeneratedProtocolMessageType('IfDynamicInfo', (_message.Message,), {
        'DESCRIPTOR' : _IFM_INTERFACES_INTERFACE_IFDYNAMICINFO,
        '__module__' : 'huawei_ifm_pb2'
        # @@protoc_insertion_point(class_scope:huawei_ifm.Ifm.Interfaces.Interface.IfDynamicInfo)
        })
      ,

      'IfStatistics' : _reflection.GeneratedProtocolMessageType('IfStatistics', (_message.Message,), {
        'DESCRIPTOR' : _IFM_INTERFACES_INTERFACE_IFSTATISTICS,
        '__module__' : 'huawei_ifm_pb2'
        # @@protoc_insertion_point(class_scope:huawei_ifm.Ifm.Interfaces.Interface.IfStatistics)
        })
      ,
      'DESCRIPTOR' : _IFM_INTERFACES_INTERFACE,
      '__module__' : 'huawei_ifm_pb2'
      # @@protoc_insertion_point(class_scope:huawei_ifm.Ifm.Interfaces.Interface)
      })
    ,
    'DESCRIPTOR' : _IFM_INTERFACES,
    '__module__' : 'huawei_ifm_pb2'
    # @@protoc_insertion_point(class_scope:huawei_ifm.Ifm.Interfaces)
    })
  ,
  'DESCRIPTOR' : _IFM,
  '__module__' : 'huawei_ifm_pb2'
  # @@protoc_insertion_point(class_scope:huawei_ifm.Ifm)
  })
_sym_db.RegisterMessage(Ifm)
_sym_db.RegisterMessage(Ifm.Interfaces)
_sym_db.RegisterMessage(Ifm.Interfaces.Interface)
_sym_db.RegisterMessage(Ifm.Interfaces.Interface.IfClearedStat)
_sym_db.RegisterMessage(Ifm.Interfaces.Interface.IfDynamicInfo)
_sym_db.RegisterMessage(Ifm.Interfaces.Interface.IfStatistics)


# @@protoc_insertion_point(module_scope)
