# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openconfig-interfaces.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='openconfig-interfaces.proto',
  package='openconfig_interfaces',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x1bopenconfig-interfaces.proto\x12\x15openconfig_interfaces\"\xc8\x1a\n\nInterfaces\x12>\n\tinterface\x18\x01 \x03(\x0b\x32+.openconfig_interfaces.Interfaces.Interface\x1a\xf9\x19\n\tInterface\x12\x42\n\x06\x63onfig\x18\x01 \x01(\x0b\x32\x32.openconfig_interfaces.Interfaces.Interface.Config\x12G\n\thold_time\x18\x02 \x01(\x0b\x32\x34.openconfig_interfaces.Interfaces.Interface.HoldTime\x12\x0c\n\x04name\x18\x03 \x01(\t\x12@\n\x05state\x18\x04 \x01(\x0b\x32\x31.openconfig_interfaces.Interfaces.Interface.State\x12P\n\rsubinterfaces\x18\x05 \x01(\x0b\x32\x39.openconfig_interfaces.Interfaces.Interface.Subinterfaces\x1an\n\x06\x43onfig\x12\x13\n\x0b\x64\x65scription\x18\x01 \x01(\t\x12\x0f\n\x07\x65nabled\x18\x02 \x01(\x08\x12\x15\n\rloopback_mode\x18\x03 \x01(\x08\x12\x0b\n\x03mtu\x18\x04 \x01(\r\x12\x0c\n\x04name\x18\x05 \x01(\t\x12\x0c\n\x04type\x18\x06 \x01(\t\x1a\xe9\x01\n\x08HoldTime\x12K\n\x06\x63onfig\x18\x01 \x01(\x0b\x32;.openconfig_interfaces.Interfaces.Interface.HoldTime.Config\x12I\n\x05state\x18\x02 \x01(\x0b\x32:.openconfig_interfaces.Interfaces.Interface.HoldTime.State\x1a\"\n\x06\x43onfig\x12\x0c\n\x04\x64own\x18\x01 \x01(\r\x12\n\n\x02up\x18\x02 \x01(\r\x1a!\n\x05State\x12\x0c\n\x04\x64own\x18\x01 \x01(\r\x12\n\n\x02up\x18\x02 \x01(\r\x1a\xd2\x08\n\x05State\x12S\n\x0c\x61\x64min_status\x18\x01 \x01(\x0e\x32=.openconfig_interfaces.Interfaces.Interface.State.AdminStatus\x12L\n\x08\x63ounters\x18\x02 \x01(\x0b\x32:.openconfig_interfaces.Interfaces.Interface.State.Counters\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0f\n\x07\x65nabled\x18\x04 \x01(\x08\x12\x0f\n\x07ifindex\x18\x05 \x01(\r\x12\x13\n\x0blast_change\x18\x06 \x01(\x04\x12\x15\n\rloopback_mode\x18\x07 \x01(\x08\x12\x0b\n\x03mtu\x18\x08 \x01(\r\x12\x0c\n\x04name\x18\t \x01(\t\x12Q\n\x0boper_status\x18\n \x01(\x0e\x32<.openconfig_interfaces.Interfaces.Interface.State.OperStatus\x12\x0c\n\x04type\x18\x0b \x01(\t\x1a\x87\x03\n\x08\x43ounters\x12\x1b\n\x13\x63\x61rrier_transitions\x18\x01 \x01(\x04\x12\x19\n\x11in_broadcast_pkts\x18\x02 \x01(\x04\x12\x13\n\x0bin_discards\x18\x03 \x01(\x04\x12\x11\n\tin_errors\x18\x04 \x01(\x04\x12\x15\n\rin_fcs_errors\x18\x05 \x01(\x04\x12\x19\n\x11in_multicast_pkts\x18\x06 \x01(\x04\x12\x11\n\tin_octets\x18\x07 \x01(\x04\x12\x17\n\x0fin_unicast_pkts\x18\x08 \x01(\x04\x12\x19\n\x11in_unknown_protos\x18\t \x01(\x04\x12\x12\n\nlast_clear\x18\n \x01(\x04\x12\x1a\n\x12out_broadcast_pkts\x18\x0b \x01(\x04\x12\x14\n\x0cout_discards\x18\x0c \x01(\x04\x12\x12\n\nout_errors\x18\r \x01(\x04\x12\x1a\n\x12out_multicast_pkts\x18\x0e \x01(\x04\x12\x12\n\nout_octets\x18\x0f \x01(\x04\x12\x18\n\x10out_unicast_pkts\x18\x10 \x01(\x04\"i\n\x0b\x41\x64minStatus\x12\x17\n\x13\x41\x64minStatus_INVALID\x10\x00\x12\x12\n\x0e\x41\x64minStatus_UP\x10\x01\x12\x14\n\x10\x41\x64minStatus_DOWN\x10\x02\x12\x17\n\x13\x41\x64minStatus_TESTING\x10\x03\"\xd1\x01\n\nOperStatus\x12\x16\n\x12OperStatus_INVALID\x10\x00\x12\x11\n\rOperStatus_UP\x10\x01\x12\x13\n\x0fOperStatus_DOWN\x10\x02\x12\x16\n\x12OperStatus_TESTING\x10\x03\x12\x16\n\x12OperStatus_UNKNOWN\x10\x04\x12\x16\n\x12OperStatus_DORMANT\x10\x05\x12\x1a\n\x16OperStatus_NOT_PRESENT\x10\x06\x12\x1f\n\x1bOperStatus_LOWER_LAYER_DOWN\x10\x07\x1a\x8b\x0c\n\rSubinterfaces\x12\\\n\x0csubinterface\x18\x01 \x03(\x0b\x32\x46.openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface\x1a\x9b\x0b\n\x0cSubinterface\x12]\n\x06\x63onfig\x18\x01 \x01(\x0b\x32M.openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.Config\x12\r\n\x05index\x18\x02 \x01(\r\x12[\n\x05state\x18\x03 \x01(\x0b\x32L.openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State\x1a=\n\x06\x43onfig\x12\x13\n\x0b\x64\x65scription\x18\x01 \x01(\t\x12\x0f\n\x07\x65nabled\x18\x02 \x01(\x08\x12\r\n\x05index\x18\x03 \x01(\r\x1a\x80\t\n\x05State\x12n\n\x0c\x61\x64min_status\x18\x01 \x01(\x0e\x32X.openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.AdminStatus\x12g\n\x08\x63ounters\x18\x02 \x01(\x0b\x32U.openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x0f\n\x07\x65nabled\x18\x04 \x01(\x08\x12\x0f\n\x07ifindex\x18\x05 \x01(\r\x12\r\n\x05index\x18\x06 \x01(\r\x12\x13\n\x0blast_change\x18\x07 \x01(\x04\x12\x0c\n\x04name\x18\x08 \x01(\t\x12l\n\x0boper_status\x18\t \x01(\x0e\x32W.openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.OperStatus\x1a\x87\x03\n\x08\x43ounters\x12\x1b\n\x13\x63\x61rrier_transitions\x18\x01 \x01(\x04\x12\x19\n\x11in_broadcast_pkts\x18\x02 \x01(\x04\x12\x13\n\x0bin_discards\x18\x03 \x01(\x04\x12\x11\n\tin_errors\x18\x04 \x01(\x04\x12\x15\n\rin_fcs_errors\x18\x05 \x01(\x04\x12\x19\n\x11in_multicast_pkts\x18\x06 \x01(\x04\x12\x11\n\tin_octets\x18\x07 \x01(\x04\x12\x17\n\x0fin_unicast_pkts\x18\x08 \x01(\x04\x12\x19\n\x11in_unknown_protos\x18\t \x01(\x04\x12\x12\n\nlast_clear\x18\n \x01(\x04\x12\x1a\n\x12out_broadcast_pkts\x18\x0b \x01(\x04\x12\x14\n\x0cout_discards\x18\x0c \x01(\x04\x12\x12\n\nout_errors\x18\r \x01(\x04\x12\x1a\n\x12out_multicast_pkts\x18\x0e \x01(\x04\x12\x12\n\nout_octets\x18\x0f \x01(\x04\x12\x18\n\x10out_unicast_pkts\x18\x10 \x01(\x04\"i\n\x0b\x41\x64minStatus\x12\x17\n\x13\x41\x64minStatus_INVALID\x10\x00\x12\x12\n\x0e\x41\x64minStatus_UP\x10\x01\x12\x14\n\x10\x41\x64minStatus_DOWN\x10\x02\x12\x17\n\x13\x41\x64minStatus_TESTING\x10\x03\"\xd1\x01\n\nOperStatus\x12\x16\n\x12OperStatus_INVALID\x10\x00\x12\x11\n\rOperStatus_UP\x10\x01\x12\x13\n\x0fOperStatus_DOWN\x10\x02\x12\x16\n\x12OperStatus_TESTING\x10\x03\x12\x16\n\x12OperStatus_UNKNOWN\x10\x04\x12\x16\n\x12OperStatus_DORMANT\x10\x05\x12\x1a\n\x16OperStatus_NOT_PRESENT\x10\x06\x12\x1f\n\x1bOperStatus_LOWER_LAYER_DOWN\x10\x07\x62\x06proto3')
)



_INTERFACES_INTERFACE_STATE_ADMINSTATUS = _descriptor.EnumDescriptor(
  name='AdminStatus',
  full_name='openconfig_interfaces.Interfaces.Interface.State.AdminStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='AdminStatus_INVALID', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AdminStatus_UP', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AdminStatus_DOWN', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AdminStatus_TESTING', index=3, number=3,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1588,
  serialized_end=1693,
)
_sym_db.RegisterEnumDescriptor(_INTERFACES_INTERFACE_STATE_ADMINSTATUS)

_INTERFACES_INTERFACE_STATE_OPERSTATUS = _descriptor.EnumDescriptor(
  name='OperStatus',
  full_name='openconfig_interfaces.Interfaces.Interface.State.OperStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OperStatus_INVALID', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OperStatus_UP', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OperStatus_DOWN', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OperStatus_TESTING', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OperStatus_UNKNOWN', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OperStatus_DORMANT', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OperStatus_NOT_PRESENT', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OperStatus_LOWER_LAYER_DOWN', index=7, number=7,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1696,
  serialized_end=1905,
)
_sym_db.RegisterEnumDescriptor(_INTERFACES_INTERFACE_STATE_OPERSTATUS)

_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE_ADMINSTATUS = _descriptor.EnumDescriptor(
  name='AdminStatus',
  full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.AdminStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='AdminStatus_INVALID', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AdminStatus_UP', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AdminStatus_DOWN', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AdminStatus_TESTING', index=3, number=3,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1588,
  serialized_end=1693,
)
_sym_db.RegisterEnumDescriptor(_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE_ADMINSTATUS)

_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE_OPERSTATUS = _descriptor.EnumDescriptor(
  name='OperStatus',
  full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.OperStatus',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OperStatus_INVALID', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OperStatus_UP', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OperStatus_DOWN', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OperStatus_TESTING', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OperStatus_UNKNOWN', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OperStatus_DORMANT', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OperStatus_NOT_PRESENT', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OperStatus_LOWER_LAYER_DOWN', index=7, number=7,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1696,
  serialized_end=1905,
)
_sym_db.RegisterEnumDescriptor(_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE_OPERSTATUS)


_INTERFACES_INTERFACE_CONFIG = _descriptor.Descriptor(
  name='Config',
  full_name='openconfig_interfaces.Interfaces.Interface.Config',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='description', full_name='openconfig_interfaces.Interfaces.Interface.Config.description', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='enabled', full_name='openconfig_interfaces.Interfaces.Interface.Config.enabled', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='loopback_mode', full_name='openconfig_interfaces.Interfaces.Interface.Config.loopback_mode', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mtu', full_name='openconfig_interfaces.Interfaces.Interface.Config.mtu', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='openconfig_interfaces.Interfaces.Interface.Config.name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='openconfig_interfaces.Interfaces.Interface.Config.type', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=450,
  serialized_end=560,
)

_INTERFACES_INTERFACE_HOLDTIME_CONFIG = _descriptor.Descriptor(
  name='Config',
  full_name='openconfig_interfaces.Interfaces.Interface.HoldTime.Config',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='down', full_name='openconfig_interfaces.Interfaces.Interface.HoldTime.Config.down', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='up', full_name='openconfig_interfaces.Interfaces.Interface.HoldTime.Config.up', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=727,
  serialized_end=761,
)

_INTERFACES_INTERFACE_HOLDTIME_STATE = _descriptor.Descriptor(
  name='State',
  full_name='openconfig_interfaces.Interfaces.Interface.HoldTime.State',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='down', full_name='openconfig_interfaces.Interfaces.Interface.HoldTime.State.down', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='up', full_name='openconfig_interfaces.Interfaces.Interface.HoldTime.State.up', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=763,
  serialized_end=796,
)

_INTERFACES_INTERFACE_HOLDTIME = _descriptor.Descriptor(
  name='HoldTime',
  full_name='openconfig_interfaces.Interfaces.Interface.HoldTime',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='config', full_name='openconfig_interfaces.Interfaces.Interface.HoldTime.config', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='state', full_name='openconfig_interfaces.Interfaces.Interface.HoldTime.state', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_INTERFACES_INTERFACE_HOLDTIME_CONFIG, _INTERFACES_INTERFACE_HOLDTIME_STATE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=563,
  serialized_end=796,
)

_INTERFACES_INTERFACE_STATE_COUNTERS = _descriptor.Descriptor(
  name='Counters',
  full_name='openconfig_interfaces.Interfaces.Interface.State.Counters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='carrier_transitions', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.carrier_transitions', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_broadcast_pkts', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.in_broadcast_pkts', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_discards', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.in_discards', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_errors', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.in_errors', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_fcs_errors', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.in_fcs_errors', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_multicast_pkts', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.in_multicast_pkts', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_octets', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.in_octets', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_unicast_pkts', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.in_unicast_pkts', index=7,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_unknown_protos', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.in_unknown_protos', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='last_clear', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.last_clear', index=9,
      number=10, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_broadcast_pkts', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.out_broadcast_pkts', index=10,
      number=11, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_discards', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.out_discards', index=11,
      number=12, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_errors', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.out_errors', index=12,
      number=13, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_multicast_pkts', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.out_multicast_pkts', index=13,
      number=14, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_octets', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.out_octets', index=14,
      number=15, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_unicast_pkts', full_name='openconfig_interfaces.Interfaces.Interface.State.Counters.out_unicast_pkts', index=15,
      number=16, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1195,
  serialized_end=1586,
)

_INTERFACES_INTERFACE_STATE = _descriptor.Descriptor(
  name='State',
  full_name='openconfig_interfaces.Interfaces.Interface.State',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='admin_status', full_name='openconfig_interfaces.Interfaces.Interface.State.admin_status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='counters', full_name='openconfig_interfaces.Interfaces.Interface.State.counters', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='openconfig_interfaces.Interfaces.Interface.State.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='enabled', full_name='openconfig_interfaces.Interfaces.Interface.State.enabled', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ifindex', full_name='openconfig_interfaces.Interfaces.Interface.State.ifindex', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='last_change', full_name='openconfig_interfaces.Interfaces.Interface.State.last_change', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='loopback_mode', full_name='openconfig_interfaces.Interfaces.Interface.State.loopback_mode', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mtu', full_name='openconfig_interfaces.Interfaces.Interface.State.mtu', index=7,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='openconfig_interfaces.Interfaces.Interface.State.name', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='oper_status', full_name='openconfig_interfaces.Interfaces.Interface.State.oper_status', index=9,
      number=10, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='openconfig_interfaces.Interfaces.Interface.State.type', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_INTERFACES_INTERFACE_STATE_COUNTERS, ],
  enum_types=[
    _INTERFACES_INTERFACE_STATE_ADMINSTATUS,
    _INTERFACES_INTERFACE_STATE_OPERSTATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=799,
  serialized_end=1905,
)

_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_CONFIG = _descriptor.Descriptor(
  name='Config',
  full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.Config',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='description', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.Config.description', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='enabled', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.Config.enabled', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='index', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.Config.index', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2239,
  serialized_end=2300,
)

_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE_COUNTERS = _descriptor.Descriptor(
  name='Counters',
  full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='carrier_transitions', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.carrier_transitions', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_broadcast_pkts', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.in_broadcast_pkts', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_discards', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.in_discards', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_errors', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.in_errors', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_fcs_errors', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.in_fcs_errors', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_multicast_pkts', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.in_multicast_pkts', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_octets', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.in_octets', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_unicast_pkts', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.in_unicast_pkts', index=7,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='in_unknown_protos', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.in_unknown_protos', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='last_clear', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.last_clear', index=9,
      number=10, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_broadcast_pkts', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.out_broadcast_pkts', index=10,
      number=11, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_discards', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.out_discards', index=11,
      number=12, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_errors', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.out_errors', index=12,
      number=13, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_multicast_pkts', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.out_multicast_pkts', index=13,
      number=14, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_octets', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.out_octets', index=14,
      number=15, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='out_unicast_pkts', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters.out_unicast_pkts', index=15,
      number=16, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1195,
  serialized_end=1586,
)

_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE = _descriptor.Descriptor(
  name='State',
  full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='admin_status', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.admin_status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='counters', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.counters', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='enabled', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.enabled', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ifindex', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.ifindex', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='index', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.index', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='last_change', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.last_change', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.name', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='oper_status', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.oper_status', index=8,
      number=9, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE_COUNTERS, ],
  enum_types=[
    _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE_ADMINSTATUS,
    _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE_OPERSTATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2303,
  serialized_end=3455,
)

_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE = _descriptor.Descriptor(
  name='Subinterface',
  full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='config', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.config', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='index', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.index', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='state', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.state', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_CONFIG, _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2020,
  serialized_end=3455,
)

_INTERFACES_INTERFACE_SUBINTERFACES = _descriptor.Descriptor(
  name='Subinterfaces',
  full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='subinterface', full_name='openconfig_interfaces.Interfaces.Interface.Subinterfaces.subinterface', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1908,
  serialized_end=3455,
)

_INTERFACES_INTERFACE = _descriptor.Descriptor(
  name='Interface',
  full_name='openconfig_interfaces.Interfaces.Interface',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='config', full_name='openconfig_interfaces.Interfaces.Interface.config', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hold_time', full_name='openconfig_interfaces.Interfaces.Interface.hold_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='openconfig_interfaces.Interfaces.Interface.name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='state', full_name='openconfig_interfaces.Interfaces.Interface.state', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='subinterfaces', full_name='openconfig_interfaces.Interfaces.Interface.subinterfaces', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_INTERFACES_INTERFACE_CONFIG, _INTERFACES_INTERFACE_HOLDTIME, _INTERFACES_INTERFACE_STATE, _INTERFACES_INTERFACE_SUBINTERFACES, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=134,
  serialized_end=3455,
)

_INTERFACES = _descriptor.Descriptor(
  name='Interfaces',
  full_name='openconfig_interfaces.Interfaces',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='interface', full_name='openconfig_interfaces.Interfaces.interface', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_INTERFACES_INTERFACE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=55,
  serialized_end=3455,
)

_INTERFACES_INTERFACE_CONFIG.containing_type = _INTERFACES_INTERFACE
_INTERFACES_INTERFACE_HOLDTIME_CONFIG.containing_type = _INTERFACES_INTERFACE_HOLDTIME
_INTERFACES_INTERFACE_HOLDTIME_STATE.containing_type = _INTERFACES_INTERFACE_HOLDTIME
_INTERFACES_INTERFACE_HOLDTIME.fields_by_name['config'].message_type = _INTERFACES_INTERFACE_HOLDTIME_CONFIG
_INTERFACES_INTERFACE_HOLDTIME.fields_by_name['state'].message_type = _INTERFACES_INTERFACE_HOLDTIME_STATE
_INTERFACES_INTERFACE_HOLDTIME.containing_type = _INTERFACES_INTERFACE
_INTERFACES_INTERFACE_STATE_COUNTERS.containing_type = _INTERFACES_INTERFACE_STATE
_INTERFACES_INTERFACE_STATE.fields_by_name['admin_status'].enum_type = _INTERFACES_INTERFACE_STATE_ADMINSTATUS
_INTERFACES_INTERFACE_STATE.fields_by_name['counters'].message_type = _INTERFACES_INTERFACE_STATE_COUNTERS
_INTERFACES_INTERFACE_STATE.fields_by_name['oper_status'].enum_type = _INTERFACES_INTERFACE_STATE_OPERSTATUS
_INTERFACES_INTERFACE_STATE.containing_type = _INTERFACES_INTERFACE
_INTERFACES_INTERFACE_STATE_ADMINSTATUS.containing_type = _INTERFACES_INTERFACE_STATE
_INTERFACES_INTERFACE_STATE_OPERSTATUS.containing_type = _INTERFACES_INTERFACE_STATE
_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_CONFIG.containing_type = _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE
_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE_COUNTERS.containing_type = _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE
_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE.fields_by_name['admin_status'].enum_type = _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE_ADMINSTATUS
_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE.fields_by_name['counters'].message_type = _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE_COUNTERS
_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE.fields_by_name['oper_status'].enum_type = _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE_OPERSTATUS
_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE.containing_type = _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE
_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE_ADMINSTATUS.containing_type = _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE
_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE_OPERSTATUS.containing_type = _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE
_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE.fields_by_name['config'].message_type = _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_CONFIG
_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE.fields_by_name['state'].message_type = _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE
_INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE.containing_type = _INTERFACES_INTERFACE_SUBINTERFACES
_INTERFACES_INTERFACE_SUBINTERFACES.fields_by_name['subinterface'].message_type = _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE
_INTERFACES_INTERFACE_SUBINTERFACES.containing_type = _INTERFACES_INTERFACE
_INTERFACES_INTERFACE.fields_by_name['config'].message_type = _INTERFACES_INTERFACE_CONFIG
_INTERFACES_INTERFACE.fields_by_name['hold_time'].message_type = _INTERFACES_INTERFACE_HOLDTIME
_INTERFACES_INTERFACE.fields_by_name['state'].message_type = _INTERFACES_INTERFACE_STATE
_INTERFACES_INTERFACE.fields_by_name['subinterfaces'].message_type = _INTERFACES_INTERFACE_SUBINTERFACES
_INTERFACES_INTERFACE.containing_type = _INTERFACES
_INTERFACES.fields_by_name['interface'].message_type = _INTERFACES_INTERFACE
DESCRIPTOR.message_types_by_name['Interfaces'] = _INTERFACES
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Interfaces = _reflection.GeneratedProtocolMessageType('Interfaces', (_message.Message,), {

  'Interface' : _reflection.GeneratedProtocolMessageType('Interface', (_message.Message,), {

    'Config' : _reflection.GeneratedProtocolMessageType('Config', (_message.Message,), {
      'DESCRIPTOR' : _INTERFACES_INTERFACE_CONFIG,
      '__module__' : 'openconfig_interfaces_pb2'
      # @@protoc_insertion_point(class_scope:openconfig_interfaces.Interfaces.Interface.Config)
      })
    ,

    'HoldTime' : _reflection.GeneratedProtocolMessageType('HoldTime', (_message.Message,), {

      'Config' : _reflection.GeneratedProtocolMessageType('Config', (_message.Message,), {
        'DESCRIPTOR' : _INTERFACES_INTERFACE_HOLDTIME_CONFIG,
        '__module__' : 'openconfig_interfaces_pb2'
        # @@protoc_insertion_point(class_scope:openconfig_interfaces.Interfaces.Interface.HoldTime.Config)
        })
      ,

      'State' : _reflection.GeneratedProtocolMessageType('State', (_message.Message,), {
        'DESCRIPTOR' : _INTERFACES_INTERFACE_HOLDTIME_STATE,
        '__module__' : 'openconfig_interfaces_pb2'
        # @@protoc_insertion_point(class_scope:openconfig_interfaces.Interfaces.Interface.HoldTime.State)
        })
      ,
      'DESCRIPTOR' : _INTERFACES_INTERFACE_HOLDTIME,
      '__module__' : 'openconfig_interfaces_pb2'
      # @@protoc_insertion_point(class_scope:openconfig_interfaces.Interfaces.Interface.HoldTime)
      })
    ,

    'State' : _reflection.GeneratedProtocolMessageType('State', (_message.Message,), {

      'Counters' : _reflection.GeneratedProtocolMessageType('Counters', (_message.Message,), {
        'DESCRIPTOR' : _INTERFACES_INTERFACE_STATE_COUNTERS,
        '__module__' : 'openconfig_interfaces_pb2'
        # @@protoc_insertion_point(class_scope:openconfig_interfaces.Interfaces.Interface.State.Counters)
        })
      ,
      'DESCRIPTOR' : _INTERFACES_INTERFACE_STATE,
      '__module__' : 'openconfig_interfaces_pb2'
      # @@protoc_insertion_point(class_scope:openconfig_interfaces.Interfaces.Interface.State)
      })
    ,

    'Subinterfaces' : _reflection.GeneratedProtocolMessageType('Subinterfaces', (_message.Message,), {

      'Subinterface' : _reflection.GeneratedProtocolMessageType('Subinterface', (_message.Message,), {

        'Config' : _reflection.GeneratedProtocolMessageType('Config', (_message.Message,), {
          'DESCRIPTOR' : _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_CONFIG,
          '__module__' : 'openconfig_interfaces_pb2'
          # @@protoc_insertion_point(class_scope:openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.Config)
          })
        ,

        'State' : _reflection.GeneratedProtocolMessageType('State', (_message.Message,), {

          'Counters' : _reflection.GeneratedProtocolMessageType('Counters', (_message.Message,), {
            'DESCRIPTOR' : _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE_COUNTERS,
            '__module__' : 'openconfig_interfaces_pb2'
            # @@protoc_insertion_point(class_scope:openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State.Counters)
            })
          ,
          'DESCRIPTOR' : _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE_STATE,
          '__module__' : 'openconfig_interfaces_pb2'
          # @@protoc_insertion_point(class_scope:openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface.State)
          })
        ,
        'DESCRIPTOR' : _INTERFACES_INTERFACE_SUBINTERFACES_SUBINTERFACE,
        '__module__' : 'openconfig_interfaces_pb2'
        # @@protoc_insertion_point(class_scope:openconfig_interfaces.Interfaces.Interface.Subinterfaces.Subinterface)
        })
      ,
      'DESCRIPTOR' : _INTERFACES_INTERFACE_SUBINTERFACES,
      '__module__' : 'openconfig_interfaces_pb2'
      # @@protoc_insertion_point(class_scope:openconfig_interfaces.Interfaces.Interface.Subinterfaces)
      })
    ,
    'DESCRIPTOR' : _INTERFACES_INTERFACE,
    '__module__' : 'openconfig_interfaces_pb2'
    # @@protoc_insertion_point(class_scope:openconfig_interfaces.Interfaces.Interface)
    })
  ,
  'DESCRIPTOR' : _INTERFACES,
  '__module__' : 'openconfig_interfaces_pb2'
  # @@protoc_insertion_point(class_scope:openconfig_interfaces.Interfaces)
  })
_sym_db.RegisterMessage(Interfaces)
_sym_db.RegisterMessage(Interfaces.Interface)
_sym_db.RegisterMessage(Interfaces.Interface.Config)
_sym_db.RegisterMessage(Interfaces.Interface.HoldTime)
_sym_db.RegisterMessage(Interfaces.Interface.HoldTime.Config)
_sym_db.RegisterMessage(Interfaces.Interface.HoldTime.State)
_sym_db.RegisterMessage(Interfaces.Interface.State)
_sym_db.RegisterMessage(Interfaces.Interface.State.Counters)
_sym_db.RegisterMessage(Interfaces.Interface.Subinterfaces)
_sym_db.RegisterMessage(Interfaces.Interface.Subinterfaces.Subinterface)
_sym_db.RegisterMessage(Interfaces.Interface.Subinterfaces.Subinterface.Config)
_sym_db.RegisterMessage(Interfaces.Interface.Subinterfaces.Subinterface.State)
_sym_db.RegisterMessage(Interfaces.Interface.Subinterfaces.Subinterface.State.Counters)


# @@protoc_insertion_point(module_scope)
