# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

import cisco_grpc_dialout_pb2 as cisco__grpc__dialout__pb2


class gRPCMdtDialoutStub(object):
  """gRPCMdtDialout defines service used for client-side streaming pushing MdtDialoutArgs.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.MdtDialout = channel.stream_stream(
        '/mdt_dialout.gRPCMdtDialout/MdtDialout',
        request_serializer=cisco__grpc__dialout__pb2.MdtDialoutArgs.SerializeToString,
        response_deserializer=cisco__grpc__dialout__pb2.MdtDialoutArgs.FromString,
        )


class gRPCMdtDialoutServicer(object):
  """gRPCMdtDialout defines service used for client-side streaming pushing MdtDialoutArgs.
  """

  def MdtDialout(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_gRPCMdtDialoutServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'MdtDialout': grpc.stream_stream_rpc_method_handler(
          servicer.MdtDialout,
          request_deserializer=cisco__grpc__dialout__pb2.MdtDialoutArgs.FromString,
          response_serializer=cisco__grpc__dialout__pb2.MdtDialoutArgs.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'mdt_dialout.gRPCMdtDialout', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))
