# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: huawei-grpc-dialout.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='huawei-grpc-dialout.proto',
  package='huawei_dialout',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x19huawei-grpc-dialout.proto\x12\x0ehuawei_dialout\":\n\x0bserviceArgs\x12\r\n\x05ReqId\x18\x01 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\x0e\n\x06\x65rrors\x18\x03 \x01(\t2`\n\x0fgRPCDataservice\x12M\n\x0b\x64\x61taPublish\x12\x1b.huawei_dialout.serviceArgs\x1a\x1b.huawei_dialout.serviceArgs\"\x00(\x01\x30\x01\x62\x06proto3')
)




_SERVICEARGS = _descriptor.Descriptor(
  name='serviceArgs',
  full_name='huawei_dialout.serviceArgs',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ReqId', full_name='huawei_dialout.serviceArgs.ReqId', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='huawei_dialout.serviceArgs.data', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='errors', full_name='huawei_dialout.serviceArgs.errors', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=45,
  serialized_end=103,
)

DESCRIPTOR.message_types_by_name['serviceArgs'] = _SERVICEARGS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

serviceArgs = _reflection.GeneratedProtocolMessageType('serviceArgs', (_message.Message,), {
  'DESCRIPTOR' : _SERVICEARGS,
  '__module__' : 'huawei_grpc_dialout_pb2'
  # @@protoc_insertion_point(class_scope:huawei_dialout.serviceArgs)
  })
_sym_db.RegisterMessage(serviceArgs)



_GRPCDATASERVICE = _descriptor.ServiceDescriptor(
  name='gRPCDataservice',
  full_name='huawei_dialout.gRPCDataservice',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=105,
  serialized_end=201,
  methods=[
  _descriptor.MethodDescriptor(
    name='dataPublish',
    full_name='huawei_dialout.gRPCDataservice.dataPublish',
    index=0,
    containing_service=None,
    input_type=_SERVICEARGS,
    output_type=_SERVICEARGS,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_GRPCDATASERVICE)

DESCRIPTOR.services_by_name['gRPCDataservice'] = _GRPCDATASERVICE

# @@protoc_insertion_point(module_scope)
