syntax = "proto3";
// Automatically generated by goyang https://github.com/openconfig/goyang
// compiled 2017-08-10T11:15:34Z
// do not delete the next line
// goyang-version 1
// module "huawei-ifm"
// revision "2017-03-23"
// revision "2013-01-01"
// namespace "http://www.huawei.com/netconf/vrp/huawei-ifm"

package huawei_ifm;

message Ifm {
  message Interfaces {
    message Interface {
      enum IfAdminStatus {
        IfAdminStatus_DOWN = 0;
        IfAdminStatus_UP = 1;
      };
      IfAdminStatus ifAdminStatus = 5;
      message IfClearedStat {
        uint64 sendByte = 29;
        uint64 sendPacket = 36;
      }
      IfClearedStat ifClearedStat = 9;
      message IfDynamicInfo {
        enum IfOperStatus {
          IfOperStatus_DOWN = 0;
          IfOperStatus_UP = 1;
        };
        IfOperStatus ifOperStatus = 5;
      }
      IfDynamicInfo ifDynamicInfo = 13;
      uint32 ifIndex = 16;
      string ifName = 25;
      message IfStatistics {
        uint64 rcvBroadPacket = 2;
        uint64 rcvDropPacket = 3;
        uint64 rcvErrorPacket = 4;
        uint64 rcvMutiPacket = 5;
        uint64 rcvUniPacket = 6;
        uint64 receiveByte = 7;
        uint64 receivePacket = 8;
        uint64 sendBroadPacket = 9;
        uint64 sendByte = 10;
        uint64 sendDropPacket = 11;
        uint64 sendErrorPacket = 12;
        uint64 sendMutiPacket = 13;
        uint64 sendPacket = 14;
        uint64 sendUniPacket = 15;
      }
      IfStatistics ifStatistics = 35;
    }
    repeated Interface interface = 1;
  }
  Interfaces interfaces = 5;
}
// Do not delete the lines below, they preserve tag information for goyang.
// goyang-tag Ifm interfaces/Interfaces 5
// goyang-tag Ifm_Interface_IfClearedStat sendByte/uint64 29
// goyang-tag Ifm_Interface_IfClearedStat sendPacket/uint64 36
// goyang-tag Ifm_Interface_IfDynamicInfo ifOperStatus/IfOperStatus 5
// goyang-tag Ifm_Interface_IfStatistics ethPortErrSts/EthPortErrSts 1
// goyang-tag Ifm_Interface_IfStatistics rcvBroadPacket/uint64 2
// goyang-tag Ifm_Interface_IfStatistics rcvMutiPacket/uint64 5
// goyang-tag Ifm_Interface_IfStatistics rcvUniPacket/uint64 6
// goyang-tag Ifm_Interface_IfStatistics receiveByte/uint64 7
// goyang-tag Ifm_Interface_IfStatistics receivePacket/uint64 8
// goyang-tag Ifm_Interface_IfStatistics sendBroadPacket/uint64 9
// goyang-tag Ifm_Interface_IfStatistics sendByte/uint64 10
// goyang-tag Ifm_Interface_IfStatistics sendMutiPacket/uint64 13
// goyang-tag Ifm_Interface_IfStatistics sendPacket/uint64 14
// goyang-tag Ifm_Interface_IfStatistics sendUniPacket/uint64 15
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts rxAlignment/uint64 1
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts rxCRC/uint64 2
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts rxFragment/uint64 3
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts rxInrangeLen/uint64 4
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts rxJabber/uint64 5
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts rxJomboOctets/uint64 6
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts rxLong/uint64 7
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts rxOverrun/uint64 8
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts rxPause/uint64 9
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts rxSymbol/uint64 10
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts rxUndersize/uint64 11
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts txJomboOctets/uint64 12
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts txLost/uint64 13
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts txOverrun/uint64 14
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts txPause/uint64 15
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts txSystem/uint64 16
// goyang-tag Ifm_Interface_IfStatistics_EthPortErrSts txUnderrun/uint64 17
// goyang-tag Ifm_Interfaces interface/Interface[] 1
// goyang-tag Ifm_Interfaces_Interface ifAdminStatus/IfAdminStatus 5
// goyang-tag Ifm_Interfaces_Interface ifClearedStat/IfClearedStat 9
// goyang-tag Ifm_Interfaces_Interface ifDynamicInfo/IfDynamicInfo 13
// goyang-tag Ifm_Interfaces_Interface ifIndex/uint32 16
// goyang-tag Ifm_Interfaces_Interface ifName/string 25
// goyang-tag Ifm_Interfaces_Interface ifStatistics/IfStatistics 35
