# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: huawei-devm.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='huawei-devm.proto',
  package='huawei_devm',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x11huawei-devm.proto\x12\x0bhuawei_devm\"\xc6\n\n\x04\x44\x65vm\x12,\n\x08\x63puInfos\x18\x05 \x01(\x0b\x32\x1a.huawei_devm.Devm.CpuInfos\x12\x38\n\x0e\x65thPortStaStss\x18\x07 \x01(\x0b\x32 .huawei_devm.Devm.EthPortStaStss\x12\x32\n\x0bmemoryInfos\x18\x0f \x01(\x0b\x32\x1d.huawei_devm.Devm.MemoryInfos\x12&\n\x05ports\x18\x14 \x01(\x0b\x32\x17.huawei_devm.Devm.Ports\x1a\xcd\x01\n\x08\x43puInfos\x12\x33\n\x07\x63puInfo\x18\x01 \x03(\x0b\x32\".huawei_devm.Devm.CpuInfos.CpuInfo\x1a\x8b\x01\n\x07\x43puInfo\x12\x10\n\x08\x65ntIndex\x18\x01 \x01(\r\x12\x10\n\x08interval\x18\x02 \x01(\r\x12\x17\n\x0fovloadThreshold\x18\x03 \x01(\r\x12\x10\n\x08position\x18\x04 \x01(\t\x12\x16\n\x0esystemCpuUsage\x18\x05 \x01(\r\x12\x19\n\x11unovloadThreshold\x18\x06 \x01(\r\x1a\xca\x01\n\x0e\x45thPortStaStss\x12\x45\n\rethPortStaSts\x18\x01 \x03(\x0b\x32..huawei_devm.Devm.EthPortStaStss.EthPortStaSts\x1aq\n\rEthPortStaSts\x12\x0e\n\x06ifName\x18\x01 \x01(\r\x12\x13\n\x0breceiveByte\x18\x02 \x01(\x04\x12\x15\n\rreceivePacket\x18\x03 \x01(\x04\x12\x10\n\x08sendByte\x18\x04 \x01(\x04\x12\x12\n\nsendPacket\x18\x05 \x01(\x04\x1a\xcd\x03\n\x0bMemoryInfos\x12<\n\nmemoryInfo\x18\x01 \x03(\x0b\x32(.huawei_devm.Devm.MemoryInfos.MemoryInfo\x1a\xff\x02\n\nMemoryInfo\x12\x14\n\x0c\x64oMemoryFree\x18\x01 \x01(\r\x12\x15\n\rdoMemoryTotal\x18\x02 \x01(\r\x12\x15\n\rdoMemoryUsage\x18\x03 \x01(\r\x12\x13\n\x0b\x64oMemoryUse\x18\x04 \x01(\r\x12\x10\n\x08\x65ntIndex\x18\x05 \x01(\r\x12\x14\n\x0cosMemoryFree\x18\x06 \x01(\r\x12\x15\n\rosMemoryTotal\x18\x07 \x01(\r\x12\x15\n\rosMemoryUsage\x18\x08 \x01(\r\x12\x13\n\x0bosMemoryUse\x18\t \x01(\r\x12\x17\n\x0fovloadThreshold\x18\n \x01(\r\x12\x10\n\x08position\x18\x0b \x01(\t\x12\x18\n\x10simpleMemoryFree\x18\x0c \x01(\r\x12\x19\n\x11simpleMemoryTotal\x18\r \x01(\r\x12\x19\n\x11simpleMemoryUsage\x18\x0e \x01(\r\x12\x17\n\x0fsimpleMemoryUse\x18\x0f \x01(\r\x12\x19\n\x11unovloadThreshold\x18\x10 \x01(\r\x1a\x8c\x02\n\x05Ports\x12*\n\x04port\x18\x01 \x03(\x0b\x32\x1c.huawei_devm.Devm.Ports.Port\x1a\xd6\x01\n\x04Port\x12=\n\x0bopticalInfo\x18\x08 \x01(\x0b\x32(.huawei_devm.Devm.Ports.Port.OpticalInfo\x1a\x8e\x01\n\x0bOpticalInfo\x12\x13\n\x0b\x62iasCurrent\x18\x01 \x01(\t\x12\x10\n\x08manuDate\x18\x02 \x01(\t\x12\x10\n\x08position\x18\x0f \x01(\t\x12\x0f\n\x07rxPower\x18\x10 \x01(\t\x12\x13\n\x0btemperature\x18\x11 \x01(\t\x12\x0f\n\x07txPower\x18\x14 \x01(\t\x12\x0f\n\x07voltage\x18\x17 \x01(\tb\x06proto3')
)




_DEVM_CPUINFOS_CPUINFO = _descriptor.Descriptor(
  name='CpuInfo',
  full_name='huawei_devm.Devm.CpuInfos.CpuInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='entIndex', full_name='huawei_devm.Devm.CpuInfos.CpuInfo.entIndex', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='interval', full_name='huawei_devm.Devm.CpuInfos.CpuInfo.interval', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ovloadThreshold', full_name='huawei_devm.Devm.CpuInfos.CpuInfo.ovloadThreshold', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='position', full_name='huawei_devm.Devm.CpuInfos.CpuInfo.position', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='systemCpuUsage', full_name='huawei_devm.Devm.CpuInfos.CpuInfo.systemCpuUsage', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unovloadThreshold', full_name='huawei_devm.Devm.CpuInfos.CpuInfo.unovloadThreshold', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=306,
  serialized_end=445,
)

_DEVM_CPUINFOS = _descriptor.Descriptor(
  name='CpuInfos',
  full_name='huawei_devm.Devm.CpuInfos',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='cpuInfo', full_name='huawei_devm.Devm.CpuInfos.cpuInfo', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_DEVM_CPUINFOS_CPUINFO, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=240,
  serialized_end=445,
)

_DEVM_ETHPORTSTASTSS_ETHPORTSTASTS = _descriptor.Descriptor(
  name='EthPortStaSts',
  full_name='huawei_devm.Devm.EthPortStaStss.EthPortStaSts',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ifName', full_name='huawei_devm.Devm.EthPortStaStss.EthPortStaSts.ifName', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='receiveByte', full_name='huawei_devm.Devm.EthPortStaStss.EthPortStaSts.receiveByte', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='receivePacket', full_name='huawei_devm.Devm.EthPortStaStss.EthPortStaSts.receivePacket', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sendByte', full_name='huawei_devm.Devm.EthPortStaStss.EthPortStaSts.sendByte', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sendPacket', full_name='huawei_devm.Devm.EthPortStaStss.EthPortStaSts.sendPacket', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=537,
  serialized_end=650,
)

_DEVM_ETHPORTSTASTSS = _descriptor.Descriptor(
  name='EthPortStaStss',
  full_name='huawei_devm.Devm.EthPortStaStss',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ethPortStaSts', full_name='huawei_devm.Devm.EthPortStaStss.ethPortStaSts', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_DEVM_ETHPORTSTASTSS_ETHPORTSTASTS, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=448,
  serialized_end=650,
)

_DEVM_MEMORYINFOS_MEMORYINFO = _descriptor.Descriptor(
  name='MemoryInfo',
  full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='doMemoryFree', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.doMemoryFree', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='doMemoryTotal', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.doMemoryTotal', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='doMemoryUsage', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.doMemoryUsage', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='doMemoryUse', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.doMemoryUse', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='entIndex', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.entIndex', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='osMemoryFree', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.osMemoryFree', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='osMemoryTotal', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.osMemoryTotal', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='osMemoryUsage', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.osMemoryUsage', index=7,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='osMemoryUse', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.osMemoryUse', index=8,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ovloadThreshold', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.ovloadThreshold', index=9,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='position', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.position', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='simpleMemoryFree', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.simpleMemoryFree', index=11,
      number=12, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='simpleMemoryTotal', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.simpleMemoryTotal', index=12,
      number=13, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='simpleMemoryUsage', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.simpleMemoryUsage', index=13,
      number=14, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='simpleMemoryUse', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.simpleMemoryUse', index=14,
      number=15, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unovloadThreshold', full_name='huawei_devm.Devm.MemoryInfos.MemoryInfo.unovloadThreshold', index=15,
      number=16, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=731,
  serialized_end=1114,
)

_DEVM_MEMORYINFOS = _descriptor.Descriptor(
  name='MemoryInfos',
  full_name='huawei_devm.Devm.MemoryInfos',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='memoryInfo', full_name='huawei_devm.Devm.MemoryInfos.memoryInfo', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_DEVM_MEMORYINFOS_MEMORYINFO, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=653,
  serialized_end=1114,
)

_DEVM_PORTS_PORT_OPTICALINFO = _descriptor.Descriptor(
  name='OpticalInfo',
  full_name='huawei_devm.Devm.Ports.Port.OpticalInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='biasCurrent', full_name='huawei_devm.Devm.Ports.Port.OpticalInfo.biasCurrent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='manuDate', full_name='huawei_devm.Devm.Ports.Port.OpticalInfo.manuDate', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='position', full_name='huawei_devm.Devm.Ports.Port.OpticalInfo.position', index=2,
      number=15, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rxPower', full_name='huawei_devm.Devm.Ports.Port.OpticalInfo.rxPower', index=3,
      number=16, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='temperature', full_name='huawei_devm.Devm.Ports.Port.OpticalInfo.temperature', index=4,
      number=17, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='txPower', full_name='huawei_devm.Devm.Ports.Port.OpticalInfo.txPower', index=5,
      number=20, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='voltage', full_name='huawei_devm.Devm.Ports.Port.OpticalInfo.voltage', index=6,
      number=23, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1243,
  serialized_end=1385,
)

_DEVM_PORTS_PORT = _descriptor.Descriptor(
  name='Port',
  full_name='huawei_devm.Devm.Ports.Port',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='opticalInfo', full_name='huawei_devm.Devm.Ports.Port.opticalInfo', index=0,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_DEVM_PORTS_PORT_OPTICALINFO, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1171,
  serialized_end=1385,
)

_DEVM_PORTS = _descriptor.Descriptor(
  name='Ports',
  full_name='huawei_devm.Devm.Ports',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='port', full_name='huawei_devm.Devm.Ports.port', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_DEVM_PORTS_PORT, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1117,
  serialized_end=1385,
)

_DEVM = _descriptor.Descriptor(
  name='Devm',
  full_name='huawei_devm.Devm',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='cpuInfos', full_name='huawei_devm.Devm.cpuInfos', index=0,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ethPortStaStss', full_name='huawei_devm.Devm.ethPortStaStss', index=1,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='memoryInfos', full_name='huawei_devm.Devm.memoryInfos', index=2,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ports', full_name='huawei_devm.Devm.ports', index=3,
      number=20, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_DEVM_CPUINFOS, _DEVM_ETHPORTSTASTSS, _DEVM_MEMORYINFOS, _DEVM_PORTS, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=35,
  serialized_end=1385,
)

_DEVM_CPUINFOS_CPUINFO.containing_type = _DEVM_CPUINFOS
_DEVM_CPUINFOS.fields_by_name['cpuInfo'].message_type = _DEVM_CPUINFOS_CPUINFO
_DEVM_CPUINFOS.containing_type = _DEVM
_DEVM_ETHPORTSTASTSS_ETHPORTSTASTS.containing_type = _DEVM_ETHPORTSTASTSS
_DEVM_ETHPORTSTASTSS.fields_by_name['ethPortStaSts'].message_type = _DEVM_ETHPORTSTASTSS_ETHPORTSTASTS
_DEVM_ETHPORTSTASTSS.containing_type = _DEVM
_DEVM_MEMORYINFOS_MEMORYINFO.containing_type = _DEVM_MEMORYINFOS
_DEVM_MEMORYINFOS.fields_by_name['memoryInfo'].message_type = _DEVM_MEMORYINFOS_MEMORYINFO
_DEVM_MEMORYINFOS.containing_type = _DEVM
_DEVM_PORTS_PORT_OPTICALINFO.containing_type = _DEVM_PORTS_PORT
_DEVM_PORTS_PORT.fields_by_name['opticalInfo'].message_type = _DEVM_PORTS_PORT_OPTICALINFO
_DEVM_PORTS_PORT.containing_type = _DEVM_PORTS
_DEVM_PORTS.fields_by_name['port'].message_type = _DEVM_PORTS_PORT
_DEVM_PORTS.containing_type = _DEVM
_DEVM.fields_by_name['cpuInfos'].message_type = _DEVM_CPUINFOS
_DEVM.fields_by_name['ethPortStaStss'].message_type = _DEVM_ETHPORTSTASTSS
_DEVM.fields_by_name['memoryInfos'].message_type = _DEVM_MEMORYINFOS
_DEVM.fields_by_name['ports'].message_type = _DEVM_PORTS
DESCRIPTOR.message_types_by_name['Devm'] = _DEVM
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Devm = _reflection.GeneratedProtocolMessageType('Devm', (_message.Message,), {

  'CpuInfos' : _reflection.GeneratedProtocolMessageType('CpuInfos', (_message.Message,), {

    'CpuInfo' : _reflection.GeneratedProtocolMessageType('CpuInfo', (_message.Message,), {
      'DESCRIPTOR' : _DEVM_CPUINFOS_CPUINFO,
      '__module__' : 'huawei_devm_pb2'
      # @@protoc_insertion_point(class_scope:huawei_devm.Devm.CpuInfos.CpuInfo)
      })
    ,
    'DESCRIPTOR' : _DEVM_CPUINFOS,
    '__module__' : 'huawei_devm_pb2'
    # @@protoc_insertion_point(class_scope:huawei_devm.Devm.CpuInfos)
    })
  ,

  'EthPortStaStss' : _reflection.GeneratedProtocolMessageType('EthPortStaStss', (_message.Message,), {

    'EthPortStaSts' : _reflection.GeneratedProtocolMessageType('EthPortStaSts', (_message.Message,), {
      'DESCRIPTOR' : _DEVM_ETHPORTSTASTSS_ETHPORTSTASTS,
      '__module__' : 'huawei_devm_pb2'
      # @@protoc_insertion_point(class_scope:huawei_devm.Devm.EthPortStaStss.EthPortStaSts)
      })
    ,
    'DESCRIPTOR' : _DEVM_ETHPORTSTASTSS,
    '__module__' : 'huawei_devm_pb2'
    # @@protoc_insertion_point(class_scope:huawei_devm.Devm.EthPortStaStss)
    })
  ,

  'MemoryInfos' : _reflection.GeneratedProtocolMessageType('MemoryInfos', (_message.Message,), {

    'MemoryInfo' : _reflection.GeneratedProtocolMessageType('MemoryInfo', (_message.Message,), {
      'DESCRIPTOR' : _DEVM_MEMORYINFOS_MEMORYINFO,
      '__module__' : 'huawei_devm_pb2'
      # @@protoc_insertion_point(class_scope:huawei_devm.Devm.MemoryInfos.MemoryInfo)
      })
    ,
    'DESCRIPTOR' : _DEVM_MEMORYINFOS,
    '__module__' : 'huawei_devm_pb2'
    # @@protoc_insertion_point(class_scope:huawei_devm.Devm.MemoryInfos)
    })
  ,

  'Ports' : _reflection.GeneratedProtocolMessageType('Ports', (_message.Message,), {

    'Port' : _reflection.GeneratedProtocolMessageType('Port', (_message.Message,), {

      'OpticalInfo' : _reflection.GeneratedProtocolMessageType('OpticalInfo', (_message.Message,), {
        'DESCRIPTOR' : _DEVM_PORTS_PORT_OPTICALINFO,
        '__module__' : 'huawei_devm_pb2'
        # @@protoc_insertion_point(class_scope:huawei_devm.Devm.Ports.Port.OpticalInfo)
        })
      ,
      'DESCRIPTOR' : _DEVM_PORTS_PORT,
      '__module__' : 'huawei_devm_pb2'
      # @@protoc_insertion_point(class_scope:huawei_devm.Devm.Ports.Port)
      })
    ,
    'DESCRIPTOR' : _DEVM_PORTS,
    '__module__' : 'huawei_devm_pb2'
    # @@protoc_insertion_point(class_scope:huawei_devm.Devm.Ports)
    })
  ,
  'DESCRIPTOR' : _DEVM,
  '__module__' : 'huawei_devm_pb2'
  # @@protoc_insertion_point(class_scope:huawei_devm.Devm)
  })
_sym_db.RegisterMessage(Devm)
_sym_db.RegisterMessage(Devm.CpuInfos)
_sym_db.RegisterMessage(Devm.CpuInfos.CpuInfo)
_sym_db.RegisterMessage(Devm.EthPortStaStss)
_sym_db.RegisterMessage(Devm.EthPortStaStss.EthPortStaSts)
_sym_db.RegisterMessage(Devm.MemoryInfos)
_sym_db.RegisterMessage(Devm.MemoryInfos.MemoryInfo)
_sym_db.RegisterMessage(Devm.Ports)
_sym_db.RegisterMessage(Devm.Ports.Port)
_sym_db.RegisterMessage(Devm.Ports.Port.OpticalInfo)


# @@protoc_insertion_point(module_scope)
