/*
    pmacct (Promiscuous mode IP Accounting package)
    pmacct is Copyright (C) 2003-2018 by Paolo Lucente
*/

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if no, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

const struct _map_dictionary_line tee_recvs_map_dictionary[] = {
  {"id", tee_recvs_map_id_handler},
  {"ip", tee_recvs_map_ip_handler},
  {"tag", tee_recvs_map_tag_handler},
  {"balance-alg", tee_recvs_map_balance_alg_handler},
  {"src_port", tee_recvs_map_src_port_handler},
#ifdef WITH_KAFKA
  {"kafka_broker", tee_recvs_map_kafka_broker_handler},
  {"kafka_topic", tee_recvs_map_kafka_topic_handler},
#endif
#ifdef WITH_ZMQ
  {"zmq_address", tee_recvs_map_zmq_address_handler},
#endif
  {"", NULL}
};
